﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;

namespace MongoExpressionConvert
{
    public static class ExpressionHelpers
    {
        //the expression targets a property that is an enumerable. It convers to x=>x.Property.Contains(value)
        public static Expression<Func<T, bool>> ConvertToListPropertyContains<T, Tkey>(Expression<Func<T, List<Tkey>>> expression, Tkey value)
        {
            //a helper method that extracts the path typed : from x=> x.Book.Id it returns Book.Id
            string propertyPath = PropertyPathHelper.GetFullPropertyPath(expression);

            if (string.IsNullOrWhiteSpace(propertyPath))
            {
                throw new Exception($"Could not get the property path from the expression {expression}");
            }
            //list of nested properties like ["Book", "Id"]
            string[] properties = propertyPath.Split(".");
            //expression start : x=>
            ParameterExpression parameter = Expression.Parameter(typeof(T), "x");

            MemberExpression memberExpression = null;

            //iterating through the properties list and creating the member expression
            foreach (string propName in properties)
            {
                if (memberExpression == null)
                {    
                    //composing : x=>x.Book.
                    memberExpression = Expression.PropertyOrField(parameter, propName);
                }
                else
                {
                    memberExpression = Expression.PropertyOrField(memberExpression, propName);
                }
            }

            MethodInfo methodInfo = typeof(List<Tkey>).GetMethod("Contains");
            //creating the equal condition : x=>x.Book.Ids.Contains(somevalue)
            MethodCallExpression body = Expression.Call(memberExpression, methodInfo, Expression.Constant(value));
            //wrapping up
            Expression<Func<T, bool>> lambaExpression = Expression.Lambda<Func<T, bool>>(body, parameter);

            return lambaExpression;
        }

        public static Expression<Func<T, bool>> ConvertToEqual<T, Tkey>(Expression<Func<T, Tkey>> expression, Tkey value)
        {
            //a helper method that extracts the path typed : from x=> x.Book.Id it returns Book.Id
            string propertyPath = PropertyPathHelper.GetFullPropertyPath(expression);

            if (string.IsNullOrWhiteSpace(propertyPath))
            {
                throw new Exception($"Could not get the property path from the expression {expression}");
            }
            //list of nested properties like ["Book", "Id"]
            string[] properties = propertyPath.Split(".");

            //expression start : x=>
            var parameter = Expression.Parameter(typeof(T), "x");

            MemberExpression memberExpression = null;

            //iterating through the properties list and creating the member expression
            foreach (string propName in properties)
            {
                if (memberExpression == null)
                {
                    //composing : x=>x.Book.
                    memberExpression = Expression.PropertyOrField(parameter, propName);
                }
                else
                {
                    memberExpression = Expression.PropertyOrField(memberExpression, propName);
                }
            }
            //creating the equal condition : x=>x.Book.Id == someValue
            BinaryExpression body = Expression.Equal(memberExpression, Expression.Constant(value));

            //wrapping up
            Expression<Func<T, bool>> lambaExpression = Expression.Lambda<Func<T, bool>>(body, parameter);

            return lambaExpression;
        }
    }
}
